/* Duplicate Tab Extension, by Twanno.
 *
 * The duplicateInTab, duplicateInWindow, waitForSessionHistory, copyTabHistory and cloneTabHistory
 * functions have been taken from Pike's Clone Window 0.2.2 extension, and have been adjusted to my needs.
 * Pike's extension: http://www.pikey.me.uk/mozilla/#cw
 */

var duplicateTab = {
  version: '1.0.2',
  prefService: null,
  prefBranch: null,
  stringBundle: null,
  tablibUsed: false,
  tablibPref: null,
  doSetScrollPosition: true,
  loadURL: null,

  handleLinkClick: null,
  gotoHistoryIndex: null,
  BrowserForward: null,
  BrowserBack: null,

  Consts: {
    TABLIB_PREFS: ['tablib.tabs_open_relative', 'extensions.tabmix.openTabNext', 'extensions.minit.open_tab', 'minit.open_tabs_relative']
  },

  getPreference: function(prefName, defaultVal) {
    var type = duplicateTab.prefService.getPrefType(prefName);
    switch (type) {
      case duplicateTab.prefService.PREF_INVALID:
        return defaultVal;
      case duplicateTab.prefService.PREF_STRING:
        return duplicateTab.prefService.getCharPref(prefName);
      case duplicateTab.prefService.PREF_INT:
        return duplicateTab.prefService.getIntPref(prefName);
      case duplicateTab.prefService.PREF_BOOL:
        return duplicateTab.prefService.getBoolPref(prefName);
      default:
        return defaultVal;
    }
    return defaultVal;
  },

  // Setup object properties
  init: function()
  {
    // load preference service
    duplicateTab.prefService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(null);
    duplicateTab.prefBranch = duplicateTab.prefService.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
    duplicateTab.prefBranch.addObserver("duplicatetab.contextMenuItems.position", duplicateTab.contextMenu.positionListener, false);

    // load stringbundle
    duplicateTab.stringBundle = document.getElementById("duplicatetab-string-bundle");

    // check whether Tab Clicking Options is installed
    duplicateTab.mergeWindows.checkTabClicking();

    // add clicking event listener to tabbar to notify selecting of a tab for merging.
    gBrowser.mStrip.childNodes[2].addEventListener("click", duplicateTab.mergeWindows.tabBarClick, false);

    // register notifications for disabling the Detach Tab toolbar button and context menu item
    gBrowser.mTabContainer.addEventListener("DOMNodeInserted", duplicateTab.mergeWindows.detachTabListener, false);
    gBrowser.mTabContainer.addEventListener("DOMNodeRemoved", duplicateTab.mergeWindows.detachTabListener, false);
    // register notification for when a toolbarbutton is added to the toolbar. This to circumvent bug 309953.
    var toolbox = document.getElementById('navigator-toolbox');
    toolbox.addEventListener("DOMNodeInserted", duplicateTab.mergeWindows.detachTabButtonListener, false);

    // add context menu items to the tab container element
    duplicateTab.contextMenu.addTabContextMenu();

    //load tablib
    if (window.tablib != undefined) {
      tablib.init();
      duplicateTab.tablibUsed = true;
      for (var i = 0; i < duplicateTab.Consts.TABLIB_PREFS.length; i++) {
        if (gBrowser.addTab.toString().indexOf(duplicateTab.Consts.TABLIB_PREFS[i]) != -1)
          duplicateTab.tablibPref = duplicateTab.Consts.TABLIB_PREFS[i];
      }
    }

    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.addObserver(duplicateTab.newWindowObserver, "duplicatetab:new-window-set-cloned-data", false);
    os.addObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-tab", false);
    os.addObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-window", false);
    os.addObserver(duplicateTab.duplicateTabObserver, "duplicatetab:merge-windows", false);
    // observer for the Merge Windows command, notify if a window is opened or closed
    os.addObserver(duplicateTab.mergeWindows.newWindowObserver, "duplicatetab:browser-window-new", false);
    os.addObserver(duplicateTab.mergeWindows.newWindowObserver, "duplicatetab:browser-window-closed", false);
    os.notifyObservers(null, "duplicatetab:browser-window-new", "");

    if (gBrowser.onDragStart && !window.TMonDrop) {
      gBrowser.mStrip.addEventListener("dragdrop", function (event) {nsDragAndDrop.drop(event, duplicateTab.tabBarDNDObserver)}, true);
      gBrowser.mStrip.addEventListener("dragover", function (event) {nsDragAndDrop.dragOver(event, duplicateTab.tabBarDNDObserver)}, true);
      gBrowser.mStrip.addEventListener("dragexit", function (event) {nsDragAndDrop.dragExit(event, duplicateTab.tabBarDNDObserver)}, true);
    }

    // add support for middle clicking on a link to open the link in a duplicated tab
    duplicateTab.handleLinkClick = window.handleLinkClick;
    window.handleLinkClick = duplicateTab.newHandleLinkClick;

    // add support for middle clicking back or forward history to duplicate tab
    duplicateTab.gotoHistoryIndex = window.gotoHistoryIndex;
    window.gotoHistoryIndex = duplicateTab.newGotoHistoryIndex;

    duplicateTab.BrowserForward = window.BrowserForward;
    window.BrowserForward = duplicateTab.newBrowserForward;

    duplicateTab.BrowserBack = window.BrowserBack;
    window.BrowserBack = duplicateTab.newBrowserBack;

    // add support for Mozilla/SeaMonkey to middle click on the history drop down menu items
    var fwButton = document.getElementById('forward-button');
    var backButton = document.getElementById('back-button');
    var fwDropDownMenu = fwButton.getElementsByTagName("menupopup")[0];
    var backDropDownMenu = backButton.getElementsByTagName("menupopup")[0];

    if (fwDropDownMenu && !fwDropDownMenu.hasAttribute('onclick'))
    {
      var fwContextMenu = document.getElementById('forwardMenu');
      fwDropDownMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      fwDropDownMenu.setAttribute('oncommand', 'gotoHistoryIndex(event); event.stopPropagation();');
      fwContextMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      fwButton.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
    }
    if (backDropDownMenu && !backDropDownMenu.hasAttribute('onclick'))
    {
      var backContextMenu = document.getElementById('backMenu');
      backDropDownMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      backDropDownMenu.setAttribute('oncommand', 'gotoHistoryIndex(event); event.stopPropagation();');
      backContextMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      backButton.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
    }
    // set Mac specific selector to use Mac toolbar button icons on Firefox 2
    if (navigator.appVersion.indexOf("Mac") != -1) {
      var ind = document.getAnonymousElementByAttribute(gBrowser, "class", "tab-drop-indicator");
      ind.setAttribute("mac", "true");
      var ids = ["duplicatetab-duptab-button", "duplicatetab-duptab-win-button", "duplicatetab-detachtab-button"];
      for (var i = 0; i < ids.length; i++) {
        var button = document.getElementById(ids[i]);
        if (button)
          button.setAttribute("mac", "true");
        button = toolbox.palette.getElementsByAttribute("id", ids[i])[0];
        button.setAttribute("mac", "true");
      }
    }
  },

  deinit: function() {
    duplicateTab.prefBranch.removeObserver("duplicatetab.contextMenuItems.position", duplicateTab.contextMenu.positionListener);
    gBrowser.mStrip.childNodes[2].removeEventListener("click", duplicateTab.mergeWindows.tabBarClick, false);
    gBrowser.mStrip.childNodes[2].removeEventListener("click", duplicateTab.contextMenu.tabBarClick, false);
    gBrowser.mTabContainer.removeEventListener("DOMNodeInserted", duplicateTab.mergeWindows.detachTabListener, false);
    gBrowser.mTabContainer.removeEventListener("DOMNodeRemoved", duplicateTab.mergeWindows.detachTabListener, false);
    document.getElementById('navigator-toolbox').addEventListener("DOMNodeInserted", duplicateTab.mergeWindows.detachTabButtonListener, false);
    duplicateTab.contextMenu.deinit();
    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.notifyObservers(null, "duplicatetab:browser-window-closed", "");
    os.removeObserver(duplicateTab.mergeWindows.newWindowObserver, "duplicatetab:browser-window-new");
    os.removeObserver(duplicateTab.mergeWindows.newWindowObserver, "duplicatetab:browser-window-closed");
    os.removeObserver(duplicateTab.newWindowObserver, "duplicatetab:new-window-set-cloned-data");
    os.removeObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-tab");
    os.removeObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-window");
    os.removeObserver(duplicateTab.duplicateTabObserver, "duplicatetab:merge-windows");

    if (gBrowser.onDragStart && !window.TMonDrop) {
      gBrowser.mStrip.removeEventListener("dragdrop", function (event) {nsDragAndDrop.drop(event, duplicateTab.tabBarDNDObserver)}, true);
      gBrowser.mStrip.removeEventListener("dragover", function (event) {nsDragAndDrop.dragOver(event, duplicateTab.tabBarDNDObserver)}, true);
      gBrowser.mStrip.removeEventListener("dragexit", function (event) {nsDragAndDrop.dragExit(event, duplicateTab.tabBarDNDObserver)}, true);
    }

    // restore original functions
    window.handleLinkClick = duplicateTab.handleLinkClick;
    window.gotoHistoryIndex = duplicateTab.gotoHistoryIndex;
    window.BrowserForward = duplicateTab.BrowserForward;
    window.BrowserBack = duplicateTab.BrowserBack;
    duplicateTab.handleLinkClick = null;
    duplicateTab.gotoHistoryIndex = null;
    duplicateTab.BrowserForward = null;
    duplicateTab.BrowserBack = null;
  },

  newHandleLinkClick: function(event, href, linkNode)
  {
    var openLinkInDuplicateTab = duplicateTab.getPreference("duplicatetab.link.middleclick.open_in_duplicate_tab", false);
    var openLinkInDuplicateTabModifier = duplicateTab.getPreference("duplicatetab.link.modifierclick.open_in_duplicate_tab", false);
    if (event.button == 0) {
      if (openLinkInDuplicateTabModifier && (event.ctrlKey || event.metaKey)) {
        // open link in new tab with history
        duplicateTab.openLinkWithHistory('tab', href, event);
        event.stopPropagation();
        return true;
      }
      else if (openLinkInDuplicateTabModifier && event.shiftKey) {
        // open link in new window with history
        duplicateTab.openLinkWithHistory('window', href);
        event.stopPropagation();
        return true;
      }
    }

    if (event.button == 1) {
      var openTab = duplicateTab.getPreference("browser.tabs.opentabfor.middleclick", true);

      var eAccelModifier = ((event.ctrlKey || event.metaKey) && !event.shiftKey && !event.altKey);
      var eAccelShiftModifier = ((event.metaKey || event.ctrlKey) && event.shiftKey && !event.altKey);
      var eNoModifier = (!event.altKey && !event.metaKey && !event.ctrlKey);
      // aEvent.shiftKey is allowed, so the duplicating the link in a new tab follows the focussing
      // of the normal middle click handling (shift key inverses the selection pref for new tabs)

      if (eAccelModifier || (eNoModifier && openLinkInDuplicateTab && openTab)) {
        // open link in new tab with history
        duplicateTab.openLinkWithHistory('tab', href, event);
        event.stopPropagation();
        return true;
      }
      else if (eAccelShiftModifier || (eNoModifier && openLinkInDuplicateTab && !openTab)) {
        // open link in new window with history
        duplicateTab.openLinkWithHistory('window', href);
        event.stopPropagation();
        return true;
      }
    }

    // call original handleLinkClick when the link will not be opened in a duplicate tab
    return duplicateTab.handleLinkClick.apply(this, arguments);
  },

  newGotoHistoryIndex: function(aEvent)
  {
    var index = aEvent.target.getAttribute("index");
    if (duplicateTab.openBFInDuplicateTab(aEvent, index))
      return true;

    // call original gotoHistoryIndex when we are not duplicating the tab
    return duplicateTab.gotoHistoryIndex.apply(this, arguments);
  },

  newBrowserForward: function(aEvent)
  {
    var index = 'forward';
    if (duplicateTab.openBFInDuplicateTab(aEvent, index))
      return;

    // call original BrowserForward when we are not duplicating the tab
    duplicateTab.BrowserForward.apply(this, arguments);
  },

  newBrowserBack: function(aEvent)
  {
    var index = 'back';
    if (duplicateTab.openBFInDuplicateTab(aEvent, index))
      return;

    // call original BrowserBack when we are not duplicating the tab
    duplicateTab.BrowserBack.apply(this, arguments);
  },

  openBFInDuplicateTab: function(aEvent, aIndex)
  {
    if (!aEvent)
      return false;

    var duplicateMiddleClick = duplicateTab.getPreference("duplicatetab.bfhistory.middleclick.open_in_duplicate_tab", false);
    var duplicateModifierClick = duplicateTab.getPreference("duplicatetab.bfhistory.modifierclick.open_in_duplicate_tab", false);

    if ('button' in aEvent && aEvent.button == 1 && duplicateMiddleClick)
    {
      var index = getIndex(aIndex);

      if (index < 0)
        return false;

      duplicateTab.doSetScrollPosition = false;
      duplicateTab.duplicateInTab(gBrowser.mCurrentTab, false, index);

      return true;
    }

    if (aEvent.type == 'command' && (aEvent.shiftKey || aEvent.ctrlKey || aEvent.metaKey) && duplicateModifierClick)
    {
      var index = getIndex(aIndex);

      if (index < 0)
        return false;

      duplicateTab.doSetScrollPosition = false;

      if (aEvent.ctrlKey || aEvent.metaKey)
        duplicateTab.duplicateInTab(gBrowser.mCurrentTab, false, index);

      else if (aEvent.shiftKey)
        duplicateTab.duplicateInWindow(gBrowser.mCurrentTab, index);

      return true;
    }

    function getIndex(ix) {
      var sHistory = gBrowser.mCurrentBrowser.webNavigation.sessionHistory;
      var sHIndex = sHistory.index;
      var sHCount = sHistory.count;

      if (ix == 'back')
        ix = sHIndex - 1;
      else if (ix == 'forward')
        ix = sHIndex + 1;
      else if (isNaN(ix))
        return -1;

      if (ix < 0 || ix > sHCount - 1) {
        // this happens when middle clicking on a disabled back or forward
        // toolbar button or menu item
        return -1;
      }

      return parseInt(ix);
    }

    return false;
  },

  checkForMiddleClick: function(node, event)
  {
    if (event.button == 1) {
      var fn = new Function("event", node.getAttribute("oncommand"));
      fn.call(node, event);

      // If the middle-click was on part of a menu, close the menu.
      // (Menus close automatically with left-click but not with middle-click.)
      duplicateTab.closeMenus(event.target);
    }
  },

  closeMenus: function(node)
  {
    if ("tagName" in node) {
      if (node.namespaceURI == "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul"
          && (node.tagName == "menupopup" || node.tagName == "popup"))
        node.hidePopup();

      duplicateTab.closeMenus(node.parentNode);
    }
  },

  openLinkWithHistory: function(target, url)
  {
    // disallow invalid urls to be handled
    if (!url || !url.length || url.indexOf(" ", 0) != -1 || /^\s*javascript:/.test(url))
    {
            // show message that hyperlink contains an invalid url and can't be loaded
      var notificationTopic = "duplicatetab:duplicate-in-" + target;
      var errorimage = "chrome://global/skin/icons/Warning.png";
      var message = duplicateTab.stringBundle.getString('duplicatetab.invalid-url.warning'); // "The hyperlink contains an invalid URL, and can not be opened in a duplicate tab."
      var buttonString = duplicateTab.stringBundle.getString('duplicatetab.invalid-url.solution'); // "Duplicate the tab anyway"
      var buttonAccessKey = duplicateTab.stringBundle.getString('duplicatetab.invalid-url.accesskey'); // 'D'
      if (gBrowser.getNotificationBox) {
        // Firefox 2.0 / 3.0 / SeaMonkey 2.0
        var notificationBox = gBrowser.getNotificationBox();
        var notification = notificationBox.getNotificationWithValue(notificationTopic);
        if (notification)
          notification.label = message;
        else {
          var buttons = [{
            label: buttonString,
            accessKey: buttonAccessKey,
            popup: null,
            callback: function() {
              var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
              os.notifyObservers(window, notificationTopic, "");
            }
          }];

          const priority = notificationBox.PRIORITY_WARNING_LOW;
          notificationBox.appendNotification(message, notificationTopic, errorimage, priority, buttons);
        }
      }
      return;
    }

    this.loadURL = url;

    switch (target) {
      case 'window':
        this.duplicateInWindow(gBrowser.selectedTab);
        break;
      case 'tab':
        var oldTab = gBrowser.selectedTab;
        var newTab = gBrowser.addTab();

        if (gBrowser.moveTabTo && this.getPreference("duplicatetab.tablib.enabled", false)) {
          var oldTabIndex = (gBrowser.getTabIndex)? gBrowser.getTabIndex(oldTab): oldTab._tPos;
          gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, oldTabIndex + 1);
        }

        // follow preferences for opening a link in a duplicate tab when arguments[2] is specified
        if (arguments[2])
        {
          var loadInBackGround = this.getPreference("browser.tabs.loadInBackground", true)
          // reverse focussing of the new tab when the shiftKey was pressed
          if (arguments[2] instanceof Event && 'shiftKey' in arguments[2] && arguments[2].shiftKey)
            loadInBackGround = !loadInBackGround;
          if (!loadInBackGround)
            gBrowser.selectedTab = newTab;
        }
        else
          gBrowser.selectedTab = newTab;

        // copy data from the original tab to the new tab.
        var clonedData = this.getClonedData(oldTab, true);
        this.setClonedData(newTab, clonedData);

        gBrowser.getBrowserForTab(newTab).loadURI(this.loadURL);
        newTab.addEventListener("SSTabRestored", duplicateTab.preventScroll, false);

        this.loadURL = null;
        break;
      default:
        this.loadURL = null;
        break;
    }
  },

  preventScroll: function(aEvent)
  {
    var browser = gBrowser.getBrowserForTab(aEvent.currentTarget);
    browser.contentDocument.defaultView.scrollTo(0, 0);
    aEvent.currentTarget.removeEventListener("SSTabRestored", duplicateTab.preventScroll, false);
  },

  // Check for duplication to tab or to window in one function where the browser to be duplicated is asigned to a value
  // Argument1 aTarget: 'tab' or 'window', to which duplication must occure
  // Argument2 aTab: optional, tab of the browser to be duplicated. If aTab is ommited, the current tab is used
  duplicateTab: function(aTarget, aTab, aEvent)
  {
    if (!aTarget)
      return false;

    if (!aTab && aEvent && !(aEvent.explicitOriginalTarget.nodeName == 'toolbarbutton' ||
              (aEvent.shiftKey && (aEvent.ctrlKey || aEvent.metaKey || aEvent.altKey)))) {
      if (gBrowser.mContextTab && gBrowser.mContextTab.localName == 'tab')
        aTab = gBrowser.mContextTab;
      else if (document.popupNode && document.popupNode.localName == 'tab')
        aTab = document.popupNode;
    }
    if (!aTab)
      aTab = gBrowser.mCurrentTab;

    if (aTarget == 'tab') {
      return this.duplicateInTab(aTab);
    } else if (aTarget == 'window') {
      return this.duplicateInWindow(aTab);
    }

    return false;
  },

  duplicateInTab: function(aTab, aNewPosition, aNewHistoryIndex)
  {
    // get the data to be cloned from the original tab
    var clonedData = this.getClonedData(aTab, false, aNewHistoryIndex);

    // Open new tab
    var newTab = gBrowser.addTab();

    // give the new tab focus if prefered by the user
    if (duplicateTab.getPreference("duplicatetab.newtab.focus", true))
      gBrowser.selectedTab = newTab;

    // place the duplicate tab at the position passed as an argument to the function
    if (gBrowser.moveTabTo && typeof aNewPosition == "number")
      gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, aNewPosition);

    // place duplicate tab next to original if desired.
    else if (gBrowser.moveTabTo)
    {
      var tablibUsed = (duplicateTab.tablibPref)?duplicateTab.getPreference(duplicateTab.tablibPref, false):false;
      if (tablibUsed || duplicateTab.getPreference("duplicatetab.tablib.enabled", false))
      {
        var newTabPos = (gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab._tPos;
        var oldTabPos = (gBrowser.getTabIndex)? gBrowser.getTabIndex(aTab): aTab._tPos;
        if (newTabPos > oldTabPos)
          gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, oldTabPos + 1);
        if (newTabPos < oldTabPos)
          gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, oldTabPos);
      }
    }

    // set the copied data of the original tab into the new tab
    this.setClonedData(newTab, clonedData);

    // Select URL bar
    if (gURLBar && duplicateTab.getPreference("duplicatetab.focus-urlbar", true))
      window.setTimeout('gURLBar.select();', 100);

    return newTab;
  },

  duplicateInWindow: function(aTab, aIndex)
  {

    // get the cloned contents from the tab
    this.newWindowObserver.clonedData = this.getClonedData(aTab, Boolean(this.loadURL), aIndex);

    // remove tab if it must be detached.
    if (this.mergeWindows.detachTab) {
      gBrowser.removeTab(aTab);
      this.mergeWindows.detachTab = false;
    }

    // Open new window
    this.newWindowObserver.openedWindow = window.openDialog(getBrowserURL(), '_blank', 'chrome,all,dialog=no');

    if (typeof(this.newWindowObserver.clonedData) == "string") {
      // sessionStore waits for the session history to be available,
      // so we don't have to wait.
      this.newWindowObserver.openedWindow.addEventListener('load',
          function() {
            var os = Components.classes["@mozilla.org/observer-service;1"]
                               .getService(Components.interfaces.nsIObserverService);
            os.notifyObservers(null, "duplicatetab:new-window-set-cloned-data", "");
          }, false);
    } else {
      // Wait until session history is available in the new window
      this.newWindowObserver.openedWindow.addEventListener('load',
          function() {duplicateTab.waitForSessionHistory(10);}, false);
    }

    return this.newWindowObserver.openedWindow;
  },

  waitForSessionHistory: function(attempts)
  {
    // Test if sessionHistory exists yet
    var webNav = duplicateTab.newWindowObserver.openedWindow.getBrowser().webNavigation;
    try    {
      webNav.sessionHistory;
    }

    // webNav.sessionHistory is not yet available, try again later
    catch (err) {
      if (attempts)
        window.setTimeout(duplicateTab.waitForSessionHistory, 100, --attempts);
      return;
    }
    if ((webNav.sessionHistory == null) && attempts) {
      window.setTimeout(duplicateTab.waitForSessionHistory, 100, --attempts);
      return;
    }

    // notify the observers that session history is available in openedBrowser, so the cloned content can be added
    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.notifyObservers(null, "duplicatetab:new-window-set-cloned-data", "");
  },

  getHistoryRange: function(aIndex, aCount, aOnlyBack, aNewIndex)
  {
    var backEnabled = duplicateTab.getPreference("duplicatetab.back-history.enabled", true);
    var forwardEnabled = duplicateTab.getPreference("duplicatetab.forward-history.enabled", true);
    var forwardLength = duplicateTab.getPreference("duplicatetab.forward-history.length", 0);
    var backLength = duplicateTab.getPreference("duplicatetab.back-history.length", 0);

    var currentIndex = typeof(aNewIndex) == "number" ? aNewIndex : aIndex;
    var lastIndex = aCount - 1;
    var firstIndex;
    var finalLength;

    if (backEnabled) {
      if (backLength == 0 || backLength >= currentIndex)
        firstIndex = 0;
      else
        firstIndex = currentIndex - backLength;
    }
    else
      firstIndex = currentIndex;

    if (forwardEnabled && !aOnlyBack) {
      if (forwardLength == 0 || currentIndex + forwardLength >= lastIndex)
        finalLength = lastIndex + 1;
      else
        finalLength = currentIndex + forwardLength + 1;
    }
    else
      finalLength = currentIndex + 1;

    var finalIndex = currentIndex - firstIndex;
    var modified = firstIndex != 0 || finalIndex != aIndex || finalLength != aCount;

    return {start: firstIndex, index: finalIndex, length: finalLength, modified: modified};
  },

  /* getSessionStoreTabState
   *
   * uses SessionStore api to get the tab state from the tab to duplicate,
   * adjusting the data to follow preferences for Duplicate Tab
   * @param aTab:      the tab to be duplicated.
   * @param aOnlyBack: if the forward history shouldn't be duplicated.
   * @param aNewIndex: the index to which the history should be set (override the current index).
   * @param aExactCopy: if the history should not be modified, overrides aOnlyBack and aNewIndex.
   *
   * returns: JSON string of tabState if success, null otherwise
   * */
  getSessionStoreTabState: function(aTab, aOnlyBack, aNewIndex, aExactCopy)
  {
    var ss = null;
    if ("@mozilla.org/browser/sessionstore;1" in Components.classes) {
      ss = Components.classes["@mozilla.org/browser/sessionstore;1"]
                     .getService(Components.interfaces.nsISessionStore);
    }
    if (!ss || !ss.getTabState || !ss.setTabState)
      return null;

    // get the range of history entries which should be duplicated.
    var range = {modified: false};
    if (!aExactCopy) {
      var history = aTab.ownerDocument.defaultView.gBrowser.getBrowserForTab(aTab)
                                                  .webNavigation.sessionHistory;
      range = this.getHistoryRange(history.index, history.count, aOnlyBack, aNewIndex);
    }

    // make sure all history entries are included.
    var ssCap = duplicateTab.getPreference("sessionstore.max_tab_back_history", -1);
    if (ssCap != -1)
      duplicateTab.prefService.setIntPref("sessionstore.max_tab_back_history", -1);

    var tabState = ss.getTabState(aTab);

    // adjust the tabData to only get the history entries to be duplicated
    // to follow the Duplicate Tab preferences.
    if (range.modified) {
      try {
        var tabStateObj = Components.utils.evalInSandbox("(" + tabState + ")", new Components.utils.Sandbox("about:blank"));
        tabStateObj.index = range.index + 1;
        tabStateObj.entries = tabStateObj.entries.slice(range.start, range.start + range.length);

        var JSON = Components.classes["@mozilla.org/dom/json;1"]
                             .createInstance(Components.interfaces.nsIJSON);
        tabState = JSON.encode(tabStateObj);
      } catch(ex) {}
    }

    // restore the preference
    if (ssCap != -1)
      duplicateTab.prefService.setIntPref("sessionstore.max_tab_back_history", ssCap);

    return tabState;
  },

  // copy a sessionHistory and put it into an array
  // Argument1 originalHistory: webNavigation.sessionHistory browser history to be copied
  // returns: an array containing a copy of the history
  copyTabHistory: function(originalHistory, aOnlyBack, aNewIndex, aExactCopy)
  {
    var range = (aExactCopy) ? {start: 0, index: originalHistory.index, length: originalHistory.count} :
                  this.getHistoryRange(originalHistory.index, originalHistory.count, aOnlyBack, aNewIndex);

    var copiedHistory = new Array();
    for (var i = range.start; i < range.length; i++) {
      copiedHistory.push(originalHistory.getEntryAtIndex(i, false));
    }
    copiedHistory.index = range.index;

    return copiedHistory;
  },

  cloneHistoryEntry: function(aEntry) {
    if (!aEntry)
      return null;
    aEntry = aEntry.QueryInterface(Components.interfaces.nsISHContainer);
    var newEntry = aEntry.clone();
    newEntry = newEntry.QueryInterface(Components.interfaces.nsISHContainer);
    newEntry.loadType = Math.floor(aEntry.loadType);
    if (aEntry.childCount) {
      for (var j = 0; j < aEntry.childCount; j++) {
          var childEntry = duplicateTab.cloneHistoryEntry(aEntry.GetChildAt(j));
          if (childEntry)
            newEntry.AddChild(childEntry, j);
      }
    }
    return newEntry;
  },

  // Clone an array of history entries into a browsers webNavigation.sessionHistory
  // Argument1 webNav: The webNavigation object of a newly created browser.
  // Argument2 originalHistory: an array containing history entries from the original browser
  cloneTabHistory: function(webNav, originalHistory)
  {
    var newHistory = webNav.sessionHistory

    newHistory.QueryInterface(Components.interfaces.nsISHistoryInternal);

    // delete history entries if they are present
    if (newHistory.count > 0)
      newHistory.PurgeHistory(newHistory.count);

    for (var i = 0; i < originalHistory.length; i++) {
      var entry = originalHistory[i].QueryInterface(Components.interfaces.nsISHEntry);
      var newEntry = duplicateTab.cloneHistoryEntry(entry);
      if (newEntry)
        newHistory.addEntry(newEntry, true);
    }

    // Go to current history location
    if (originalHistory.index < originalHistory.length)
      gotoHistoryIndex(10);

    function gotoHistoryIndex(attempts) {
      try {
        webNav.gotoIndex(originalHistory.index);
      }
      catch(e) {
        // do some math to increase the timeout
        // each time we try to update the history index
        if (attempts)
          setTimeout(gotoHistoryIndex, (11 - attempts) * (15 - attempts), --attempts);
      }
    }
  },

  /* getClonedData
   * get the data from the tab which is to be cloned.
   *
   * @param aTab:       tabbrowser tab which should be cloned.
   * @param aOnlyBack:  optional, boolean. If true the forward history should not be cloned.
   * @param aNewIndex:  optional, integer. New index to set the history to.
   * @param aExactCopy: optional, boolean. If true no data should be lost
   *                      (aOnlyBack and aNewIndex have no effect when this is set)
   *
   * returns either a JSON string or an object containing the data which should be cloned.
   */
  getClonedData: function(aTab, aOnlyBack, aNewIndex, aExactCopy)
  {
    // try sessionStore first
    var tabState = this.getSessionStoreTabState(aTab, aOnlyBack, aNewIndex, aExactCopy);
    if (tabState)
      return tabState;

    var browser = aTab.ownerDocument.defaultView.gBrowser.getBrowserForTab(aTab);
    var clonedData = new Array();
    clonedData[0] = this.copyTabHistory(browser.webNavigation.sessionHistory, aOnlyBack, aNewIndex, aExactCopy);
    clonedData[1] = browser.contentWindow.scrollX;
    clonedData[2] = browser.contentWindow.scrollY;
    clonedData[3] = browser.markupDocumentViewer.textZoom;

    return clonedData;
  },

  /* getClonedData for older versions of e.g. Tab Clicking Options */
  getClonedContent: function(aBrowser, aOnlyBack, aNewIndex, aExactCopy)
  {
    var tabbrowser = aBrowser.ownerDocument.defaultView.getBrowser();
    for (var i = 0; i < tabbrowser.tabContainer.childNodes.length; i++) {
      var tab = tabbrowser.tabContainer.childNodes.item(i);
      if (tab.linkedBrowser == aBrowser)
        return this.getClonedData(tab, aOnlyBack, aNewIndex, aExactCopy);
    }
    return null;
  },

  /* setClonedData
   * sets the data cloned from the original tab into a new tab.
   * @param aTab:        the new tab the data has to be set to.
   * @param aClonedData: JSON string or object containing the data.
   *
   * returns  boolean to indicate successfullness
   */
  setClonedData: function(aTab, aClonedData)
  {
    // if aClonedData is a JSON string we can use sessionStore.
    if (typeof(aClonedData) == "string") {
      var ss = Components.classes["@mozilla.org/browser/sessionstore;1"]
                         .getService(Components.interfaces.nsISessionStore);
      ss.setTabState(aTab, aClonedData);
      return true;
    }

    var browser = aTab.ownerDocument.defaultView.gBrowser.getBrowserForTab(aTab);
    function setTextZoom(attempts) {
      browser.markupDocumentViewer.textZoom = aClonedData[3];
      if (attempts && browser.markupDocumentViewer.textZoom != aClonedData[3])
        setTimeout(setTextZoom, 10, --attempts)
    }

    function setScrollPosition(attempts) {
      browser.contentWindow.scrollTo(aClonedData[1], aClonedData[2]);
      if (attempts && (browser.contentWindow.scrollX != aClonedData[1] || browser.contentWindow.scrollY != aClonedData[2]))
        setTimeout(setScrollPosition, 10, --attempts);
    }

    if (aClonedData[0].length == 0)
      return false;

    this.cloneTabHistory(browser.webNavigation, aClonedData[0]);

    if (this.getPreference("duplicatetab.scrollposition.enabled", true) && this.doSetScrollPosition)
      setScrollPosition(15);

    if (this.getPreference("duplicatetab.zoomfactor.enabled", true))
      setTextZoom(15);

    if (!this.doSetScrollPosition)
      this.doSetScrollPosition = true;

    return true;
  },

  /* setClonedData for older versions of e.g. Tab Clicking Options */
  setClonedContent: function(aBrowser, aClonedContent)
  {
    var tabbrowser = aBrowser.ownerDocument.defaultView.getBrowser();
    for (var i = 0; i < tabbrowser.tabContainer.childNodes.length; i++) {
      var tab = tabbrowser.tabContainer.childNodes.item(i);
      if (tab.linkedBrowser == aBrowser)
        return this.setClonedData(tab, aClonedContent);
    }
    return false;
  },

  buttonObserver:
  {
    onDragOver: function (aEvent, aFlavour, aDragSession)
    {
      aEvent.target.setAttribute("dragover", "true");
      return true;
    },
    onDragExit: function (aEvent, aDragSession)
    {
      aEvent.target.removeAttribute("dragover");
    },
    onDrop: function (aEvent, aXferData, aDragSession)
    {
      var xferFlavourType = aXferData.flavour.contentType;
      var tab;
      if (xferFlavourType == 'text/x-moz-tab' || xferFlavourType == 'text/x-minit-tab') {
        tab = gBrowser.mTabs[aXferData.data];
      } else {
        var xferData = aXferData.data.split("\n");
        var url = xferData[0] || xferData[1];

        if (aDragSession.sourceNode && aDragSession.sourceNode.parentNode == gBrowser.mTabContainer) {
          tab = aDragSession.sourceNode;
        } else if (url == gBrowser.getBrowserForTab(gBrowser.mCurrentTab).currentURI.spec) {
          tab = gBrowser.mCurrentTab
        } else if (url && aEvent.target.id != "duplicatetab-detachtab-button") {
          var where = (aEvent.target.id == "duplicatetab-duptab-button") ? 'tab' : 'window';
          duplicateTab.openLinkWithHistory(where, url);
        }
      }
      if (!tab)
        return;

      switch (aEvent.target.id) {
        case "duplicatetab-duptab-button":
          duplicateTab.duplicateTab('tab', tab, aEvent);
          break;
        case "duplicatetab-duptab-win-button":
          duplicateTab.duplicateTab('window', tab, aEvent);
          break;
        case "duplicatetab-detachtab-button":
          duplicateTab.mergeWindows.detachTabs(tab, aEvent);
          break;
      }
    },
    getSupportedFlavours: function ()
    {
      var flavourSet = new FlavourSet();
      flavourSet.appendFlavour("text/x-moz-tab");
      flavourSet.appendFlavour("text/x-minit-tab");
      flavourSet.appendFlavour("text/unicode");
      flavourSet.appendFlavour("text/html");
      return flavourSet;
    }
  },

  newWindowObserver:
  {
    clonedData: null,
    openedWindow: null,

    observe: function(aSubject, aTopic, aData)
    {
      switch (aTopic) {
        case "duplicatetab:new-window-set-cloned-data":
          // only the window with this.openedWindow set will
          // respond to this notification.
          if (this.openedWindow) {
            var openedTab = this.openedWindow.getBrowser().selectedTab;

            duplicateTab.setClonedData(openedTab, this.clonedData);

            if (duplicateTab.loadURL) {
              this.openedWindow.gBrowser.getBrowserForTab(openedTab).loadURI(duplicateTab.loadURL);
              openedTab.addEventListener("SSTabRestored", duplicateTab.preventScroll, false);

              duplicateTab.loadURL = null;
            }

            if (duplicateTab.mergeWindows.detachIndex > -1)
              duplicateTab.mergeWindows.addOtherTabs(this.openedWindow);

            this.openedWindow = null;
            this.clonedData = null;
          }
          break;
      }
    }
  },

  duplicateTabObserver: {
    observe: function(aSubject, aTopic, aData) {
      // aSubject should be this window object
      if (window != aSubject)
        return;

      switch (aTopic)
      {
        case "duplicatetab:duplicate-in-tab":
          duplicateTab.duplicateTab('tab', null, null);
          break;
        case "duplicatetab:duplicate-in-window":
          duplicateTab.duplicateTab('window', null, null);
          break;
        case "duplicatetab:merge-windows":
          duplicateTab.mergeWindows.mergeWindows();
          break;
      }
    }
  },

  tabBarDNDObserver: {
    onDragOver: function (aEvent, aXferData, aDragSession)
    {
      var ind = document.getAnonymousElementByAttribute(gBrowser, "class", "tab-drop-indicator");
      if (!ind)
        return nsDragAndDrop.dragOver(aEvent, gBrowser);
      if (('ctrlKey' in aEvent || 'metaKey' in aEvent) && (aEvent.ctrlKey || aEvent.metaKey))
        ind.setAttribute('duplicatetab', 'true');
      else
        ind.removeAttribute('duplicatetab');
      return nsDragAndDrop.dragOver(aEvent, gBrowser);
    },
    onDrop: function (aEvent, aXferData, aDragSession)
    {
      if (('ctrlKey' in aEvent || 'metaKey' in aEvent) && (aEvent.ctrlKey || aEvent.metaKey)
          && aDragSession.sourceNode && aDragSession.sourceNode.parentNode == gBrowser.mTabContainer)
      {
        var newIndex = (gBrowser.getDropIndex)? gBrowser.getDropIndex(aEvent): gBrowser.getNewIndex(aEvent);
        duplicateTab.duplicateInTab(aDragSession.sourceNode, newIndex);
        return true;
      }
      else
        return nsDragAndDrop.drop(aEvent, gBrowser);
    },
    onDragExit: function (aEvent, aDragSession)
    {
      var ind = document.getAnonymousElementByAttribute(gBrowser, "class", "tab-drop-indicator");
      if (ind)
        ind.removeAttribute('duplicatetab');
      return nsDragAndDrop.dragExit(aEvent, gBrowser);
    },
    getSupportedFlavours: function ()
    {
      var flavourSet = new FlavourSet();
      flavourSet.appendFlavour("text/unicode");
      flavourSet.appendFlavour("text/html");
      return flavourSet;
    }
  },

  selectBrowser: function() {
    // maintain support for older (< 0.6.5) versions of Tab Clicking Options
    this.mergeWindows.selectBrowser(this.selectedTab);
    delete this.selectedTab
  }
}

window.addEventListener("load", duplicateTab.init, false);
window.addEventListener("unload", duplicateTab.deinit, false);
